{
  Implementation of the Basic authentication as specified in
  RFC 2616

  Copyright: (c) Chad Z. Hower and The Winshoes Working Group.


  Author: Doychin Bondzhev (doychin@dsoft-bg.com)
}

unit IdAuthentication;

interface

Uses
  Classes, IdHeaderList, IdGlobal;

Type
  TIdAuthenticationSchemes = (asBasic, asDigest, asNTLM, asUnknown);
  TIdAuthSchemeSet = set of TIdAuthenticationSchemes;

  TIdAuthWhatsNext = (wnAskTheProgram, wnDoRequest, wnFail);

  TIdAuthentication = class(TPersistent)
  protected
    FRetries: Integer;
    FUsername: String;
    FPassword: String;

    FStatus: Integer;

    FAuthParams: TIdHeaderList;

    function ReadAuthInfo(AuthName: String): String;
    function DoNext: TIdAuthWhatsNext; virtual; abstract;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    procedure SetAuthParams(AValue: TIdHeaderList);
    function Authentication: String; virtual; abstract;
    function Next: TIdAuthWhatsNext;

    property AuthParams: TIdHeaderList read FAuthParams write SetAuthParams;
    property Username: String read FUsername write FUsername;
    property Password: String read FPassword write FPassword;
                   
    property Retries: Integer read FRetries;
  end;

  TIdAuthenticationClass = class of TIdAuthentication;

  TIdBasicAuthentication = class(TIdAuthentication)
  protected
    FRealm: String;
    function DoNext: TIdAuthWhatsNext; override;
  public
    constructor Create; override;
    function Authentication: String; override;

    property Realm: String read FRealm write FRealm;
  end;

  { Support functions }
  procedure RegisterAuthenticationMethod(MethodName: String; AuthClass: TIdAuthenticationClass);
  function FindAuthClass(AuthName: String): TIdAuthenticationClass;

implementation

Uses
  IdCoderMIME, SysUtils;

Type
  TAuthListObject = class(TObject)
    Auth: TIdAuthenticationClass;
  end;

Var
  AuthList: TStringList = nil;

procedure RegisterAuthenticationMethod(MethodName: String; AuthClass: TIdAuthenticationClass);
Var
  A: TAuthListObject;
begin
  if not Assigned(AuthList) then begin
    AuthList := TStringList.Create;
  end;

  A := TAuthListObject.Create;
  A.Auth := AuthClass;
  AuthList.AddObject(MethodName, A);
end;

function FindAuthClass(AuthName: String): TIdAuthenticationClass;
begin
  if AuthList.IndexOf(AuthName) = -1 then Abort;
  result := TAuthListObject(AuthList.Objects[AuthList.IndexOf(AuthName)]).Auth;
end;

{ TIdAuthentication }

constructor TIdAuthentication.Create;
begin
  inherited Create;
  FRetries := 0;
  FStatus := 0;
end;

destructor TIdAuthentication.Destroy;
begin
  if Assigned(FAuthParams) then begin
    FAuthParams.Free;
  end;
  inherited Destroy;
end;

procedure TIdAuthentication.SetAuthParams(AValue: TIdHeaderList);
begin
  if not Assigned(FAuthParams) then begin
    FAuthParams := TIdHeaderList.Create;
  end;

  FAuthParams.Assign(AValue);
end;

function TIdAuthentication.ReadAuthInfo(AuthName: String): String;
Var
  i: Integer;
begin
  if Assigned(FAuthParams) then begin
    for i := 0 to FAuthParams.Count - 1 do begin
      if IndyPos(AuthName, FAuthParams[i]) = 1 then begin
        result := FAuthParams[i];
        exit;
      end;
    end;
  end
  else begin
    result := '';
  end;
end;

function TIdAuthentication.Next: TIdAuthWhatsNext;
begin
  result := DoNext;

  if result = wnDoRequest then begin
    Inc(FRetries);
  end;
end;

{ TIdBasicAuthentication }

constructor TIdBasicAuthentication.Create;
begin
  inherited Create;
  FStatus := 0;
end;

function TIdBasicAuthentication.Authentication: String;
begin
  result := 'Basic ' {do not localize}
    + TIdEncoderMIME.EncodeString(FUsername + ':' + FPassword);
end;

function TIdBasicAuthentication.DoNext: TIdAuthWhatsNext;
Var
  S: String;
  Params: TStringList;
begin
  result := wnDoRequest;

  S := ReadAuthInfo('Basic');
  Fetch(S);

  Params := TStringList.Create;

  while Length(S) > 0 do begin
    Params.Add(Fetch(S, ', '));
  end;

  FRealm := Copy(Params.Values['realm'], 2, Length(Params.Values['realm']) - 2);

  Params.Free;

  case FStatus of
    0: begin
      if (Length(Username) > 0) and (Length(Password) > 0) and (FRetries = 0) then begin
        result := wnDoRequest;
      end
      else begin
        result := wnAskTheProgram;
        FStatus := 1;
      end;
    end;
    1: begin
      result := wnDoRequest;
    end;
  end;
end;

initialization
  RegisterAuthenticationMethod('Basic', TIdBasicAuthentication);
finalization
  if Assigned(AuthList) then begin
    while AuthList.Count > 0 do begin
      AuthList.Objects[0].Free;
      AuthList.Delete(0);
    end;
    AuthList.Free;
  end;
end.
