unit IdThread;

interface

uses
  Classes,
  SysUtils, SyncObjs, IdGlobal;

type
  TIdThreadStopMode = (smTerminate, smSuspend);
  TIdExceptionEvent = procedure(ASender: TObject; AException: Exception) of object;
  TIdThread = class;
  TIdNotifyThreadEvent = procedure(AThread: TIdThread) of object;

  TIdSync = class(TObject)
  protected
    FThread: TIdThread;
  public
    constructor Create(AThread: TIdThread); reintroduce; virtual;
    procedure DoSynchronize; virtual; abstract;
    procedure Notify;
    procedure Synchronize;
    //
    property Thread: TIdThread read FThread;
  end;

  TIdSyncMethod = class(TIdSync)
  protected
    FMethod: TThreadMethod;
  public
    constructor Create(AMethod: TThreadMethod); reintroduce;
    procedure DoSynchronize; override;
  end;

  TIdThread = class(TThread)
  protected
    FData: TObject;
    FLock: TCriticalSection;
    FStopMode: TIdThreadStopMode;
    FStopped: Boolean;
    FTerminatingException: string;
    FOnException: TIdExceptionEvent;
    FOnStopped: TIdNotifyThreadEvent;
    //
    procedure AfterRun; virtual; // Not abstract - otherwise it is required
    procedure BeforeRun; virtual; // Not abstract - otherwise it is required
    procedure DoStopped; virtual;
    procedure Execute; override;
    function GetStopped: Boolean;
    procedure Run; virtual; abstract;
  public
    constructor Create(ACreateSuspended: Boolean = True); virtual;
    destructor Destroy; override;
    // In TIdThread to assure that at least one TIdThread does exist
    procedure Notify(ASync: TIdSync); overload;
    procedure Notify(AMethod: TThreadMethod); overload;
    procedure Start; virtual;
    procedure Stop; virtual;
    // Synchronize -  Here to expose it
    procedure Synchronize(Method: TThreadMethod);
    // Here to make virtual
    procedure Terminate; virtual;
    procedure TerminateAndWaitFor; virtual;
    //
    property Data: TObject read FData write FData;
    property StopMode: TIdThreadStopMode read FStopMode write FStopMode;
    property Stopped: Boolean read GetStopped;
    property OnException: TIdExceptionEvent read FOnException write FOnException;
    property OnStopped: TIdNotifyThreadEvent read FOnStopped write FOnStopped;
    // Terminated exposed
    property Terminated;
    property TerminatingException: string read FTerminatingException;
  end;

  TIdThreadClass = class of TIdThread;

  // This is done with a NotifyThread instead of PostMessage because starting with D6/Kylix Borland
  // radically modified the mecanisms for .Synchronize. This is a bit more code in the end, but
  // its source compatible and does not rely on Indy directly accessing any OS APIs and performance
  // is still more than acceptable, especiall considering Notifications are low priority.
  TIdNotifyThread = class(TThread)
  protected
    FEvent: TIdLocalEvent;
    FNotifications: TThreadList;
  public
    class procedure LockNotifications;
    class procedure UnlockNotifications;
    procedure AddNotification(ASync: TIdSync);
    constructor Create(ASuspended: Boolean);
    destructor Destroy; override;
    procedure Execute; override;
    // NOTE: Cant override
    procedure Terminate;
  end;

{$IFNDEF CBUILDER}
ThreadVar
  // Warning: This cannot be used outside of the package, or unless this is compiled into the EXE.
  ThreadSelf: TIdThread;
{$ENDIF}

var
  GNotifyThread: TIdNotifyThread = nil;

implementation

procedure TIdThread.TerminateAndWaitFor;
begin
  Terminate;
  FStopped := True;
  WaitFor;
end;

procedure TIdThread.AfterRun;
begin
end;

procedure TIdThread.BeforeRun;
begin
end;

procedure TIdThread.Execute;
begin
  try
    {$IFNDEF CBUILDER}
    ThreadSelf := Self;
    {$ENDIF}
    while not Terminated do begin
      if Stopped then begin
        DoStopped;
        // It is possible that either in the DoStopped or from another thread, the thread is
        // restarted, in which case we dont want to restop it.
        if Stopped then begin
          Suspended := True; // Thread manager will revive us
          if Terminated then begin
            Break;
          end;
        end;
      end;
      BeforeRun; try
        while not Stopped do begin
          Run;
        end;
      finally AfterRun; end;
    end;
  except
    on E: Exception do begin
      FTerminatingException := E.Message;
      if Assigned(FOnException) then begin
        FOnException(self, E);
      end;
      Terminate;
    end;
  end;
end;

procedure TIdThread.Synchronize(Method: TThreadMethod);
begin
  inherited Synchronize(Method);
end;

constructor TIdThread.Create(ACreateSuspended: Boolean);
begin
  // Before inherited - inherited creates the actual thread and if not suspeded will start before
  // we initialize
  FStopped := ACreateSuspended;
  FLock := TCriticalSection.Create;
  try
    inherited Create(ACreateSuspended);
  except
    FreeAndNil(FLock);
    raise;
  end;
  FLock.Enter; try
    if GNotifyThread = nil then begin
      GNotifyThread := TIdNotifyThread.Create(False);
    end;
  finally FLock.Leave; end;
end;

destructor TIdThread.Destroy;
begin
  FreeAndNil(FLock);
  FreeAndNil(FData);
  inherited Destroy;
end;

procedure TIdThread.Start;
begin
  FLock.Enter; try
    if Stopped then begin
      // Resume is also called for smTerminate as .Start can be used to initially start a
      // thread that is created suspended
      FStopped := False;
      Suspended := False;
    end;
  finally FLock.Leave; end;
end;

procedure TIdThread.Stop;
begin
  FLock.Enter; try
    if not Stopped then begin
      case FStopMode of
        smTerminate: Terminate;
        // DO NOT suspend here. Suspend is immediate. See Execute for implementation
        smSuspend: ;
      end;
      FStopped := True;
    end;
  finally FLock.Leave; end;
end;

function TIdThread.GetStopped: Boolean;
begin
  FLock.Enter; try
    // Suspended may be true if checking stopped from another thread
    Result := Terminated or FStopped or Suspended;
  finally FLock.Leave; end;
end;

procedure TIdThread.DoStopped;
begin
  if Assigned(OnStopped) then begin
    OnStopped(Self);
  end;
end;

procedure TIdThread.Notify(AMethod: TThreadMethod);
begin
  Notify(TIdSyncMethod.Create(AMethod));
end;

{ TIdSync }

constructor TIdSync.Create(AThread: TIdThread);
begin
  inherited Create;
  FThread := AThread;
end;

procedure TIdSync.Notify;
begin
  FThread.Notify(Self);
end;

procedure TIdSync.Synchronize;
begin
  FThread.Synchronize(DoSynchronize);
end;

procedure TIdThread.Notify(ASync: TIdSync);
begin
  GNotifyThread.AddNotification(ASync);
end;

{ TIdNotifyThread }

procedure TIdNotifyThread.AddNotification(ASync: TIdSync);
begin
  FNotifications.Add(ASync);
  FEvent.SetEvent;
end;

constructor TIdNotifyThread.Create(ASuspended: Boolean);
begin
  FreeOnTerminate := True;
  FNotifications := TThreadList.Create;
  FEvent := TIdLocalEvent.Create;
  // Must be before - Thread starts running when we call inherited
  inherited Create(ASuspended);
end;

destructor TIdNotifyThread.Destroy;
begin
  FreeAndNil(FEvent);
  FreeAndNil(FNotifications);
  inherited Destroy;
end;

procedure TIdNotifyThread.Execute;
var
  LNotify: TList;
  LSync: TIdSync;
begin
  while not Terminated do begin
    FEvent.WaitFor;
    LNotify := FNotifications.LockList; try
      while not Terminated and (LNotify.Count > 0) do begin
        LSync := TIdSync(LNotify.Items[0]);
        LNotify.Delete(0);
        if LSync <> nil then begin
          Synchronize(LSync.DoSynchronize);
          FreeAndNil(LSync);
        end;
      end;
    finally FNotifications.UnlockList; end;
  end;
end;

class procedure TIdNotifyThread.LockNotifications;
begin
  if Assigned(GNotifyThread) then begin
    GNotifyThread.FNotifications.LockList;
  end;
end;

procedure TIdNotifyThread.Terminate;
begin
  // Close it here so the WaitFor will kick out
  FreeAndNil(FEvent);
  inherited Terminate;
end;

class procedure TIdNotifyThread.UnlockNotifications;
begin
  if Assigned(GNotifyThread) then begin
    GNotifyThread.FNotifications.UnlockList;
  end;
end;

{ TIdSyncMethod }

constructor TIdSyncMethod.Create(AMethod: TThreadMethod);
begin
  FMethod := AMethod;
end;

procedure TIdSyncMethod.DoSynchronize;
begin
  FMethod;
end;

procedure TIdThread.Terminate;
begin
  inherited Terminate;
end;

initialization
finalization
//  GNotifyThread.Terminate;
  // Will free itself using FreeOnTerminate
//  GNotifyThread := nil;
end.
